#ifndef _OBJECTS_H_
#define _OBJECTS_H_
#include <math.h>
#include <_null.h>
#include "lights.h"
#include "vector.h"
#include "surface.h"
#include "texture.h"
#include "ray.h"

const double GeomThreshold = 0.001;

class GObject {
  public:
    SurfaceData data;
    GObject() { }
    GObject(SurfaceData data)
    {
      GObject::data = data;
      next = NULL;
    }
    virtual bool Intersect(Ray ray, double& dist) { return false; }
    virtual Vector GetColor(Vector point) { return Vector(0); }
    virtual Vector GetNormal(Vector point) { return Vector(0); }
    GObject *next;
};

class Plane: public GObject {
    double d;
    int i, j, k;
    Texture *tex;
    Vector c;
  public:
    Plane(SurfaceData data, double d, int i, Texture *tex):
    GObject(data)
    {
      Plane::d = d;
      Plane::i = i;
      switch (i)
      {
        case 0: j = 1; k = 2; break;
        case 1: j = 0; k = 2; break;
        case 2: j = 0; k = 1; break;
      }
      Plane::tex = tex;
    }
    Plane(SurfaceData data, double d, int i, Vector color):
    GObject(data)
    {
      Plane::d = d;
      Plane::i = i;
      j = k = 0;
      c = color;
      tex = NULL;
    }
    bool Intersect(Ray ray, double& dist)
    {
      if ((ray.dir[i] >= 0.0 && ray.org[i] > d) || (ray.dir[i] <= 0.0 && ray.org[i] < d))
        return false;
      dist = fabs((ray.org[i] - d) / ray.dir[i]);
      return true;
    }
    Vector GetNormal(Vector point)
    {
      Vector res(0);
      res[i] = 1.0;
      return res;
    }
    Vector GetColor(Vector p)
    {
      if (tex == NULL) return c;
      return tex -> GetColor(p[j], p[k]);
    }
};

class XYPlane: public Plane {
  public:
    XYPlane(SurfaceData data, double d, Texture *tex):
    Plane(data, d, 2, tex) { }
    XYPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 2, color) { }
};

class YZPlane: public Plane {
  public:
    YZPlane(SurfaceData data, double d, Texture *tex):
    Plane(data, d, 0, tex) { }
    YZPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 0, color) { }
};

class ZXPlane: public Plane {
  public:
    ZXPlane(SurfaceData data, double d, Texture *tex):
    Plane(data, d, 1, tex) { }
    ZXPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 1, color) { }
};

/*class Plane: public GObject {
    double d, size;
    int i, j, k;
    Vector c1, c2;
  public:
    Plane(SurfaceData data, double d, int i, Vector c1, Vector c2, double size):
    GObject(data)
    {
      Plane::d = d;
      Plane::i = i;
      j = (i + 1) % 3;
      k = (i + 2) % 3;
      Plane::c1 = c1;
      Plane::c2 = c2;
      Plane::size = size;
    }
    Plane(SurfaceData data, double d, int i, Vector c1):
    GObject(data)
    {
      Plane::d = d;
      Plane::i = i;
      j = k = 0;
      Plane::c1 = c1;
      c2 = c1;
      Plane::size = 1;
    }
    bool Intersect(Ray ray, double& dist)
    {
      if ((ray.dir[i] >= 0.0 && ray.org[i] > d) || (ray.dir[i] <= 0.0 && ray.org[i] < d))
        return false;
      dist = fabs((ray.org[i] - d) / ray.dir[i]);
      return true;
    }
    Vector GetNormal(Vector point)
    {
      Vector res(0);
      res[i] = 1.0;
      return res;
    }
    Vector GetColor(Vector p)
    {
      int a = fabs(p[j] / size), b = fabs(p[k] / size);
      if (p[j] < 0) a++;
      if (p[k] < 0) b++;
      return ((b ^ a) & 1) ? c1 : c2;
    }
};

class XYPlane: public Plane {
  public:
    XYPlane(SurfaceData data, double d, Vector c1, Vector c2, double size):
    Plane(data, d, 2, c1, c2, size) { }
    XYPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 2, color) { }
};

class YZPlane: public Plane {
  public:
    YZPlane(SurfaceData data, double d, Vector c1, Vector c2, double size):
    Plane(data, d, 0, c1, c2, size) { }
    YZPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 0, color) { }
};

class ZXPlane: public Plane {
  public:
    ZXPlane(SurfaceData data, double d, Vector c1, Vector c2, double size):
    Plane(data, d, 1, c1, c2, size) { }
    ZXPlane(SurfaceData data, double d, Vector color):
    Plane(data, d, 1, color) { }
};*/

class Sphere: public GObject {
    double r, r2;
    Texture *tex;
    Vector org, color;
  public:
    Sphere(SurfaceData data, Vector org, double r, Vector color):
    GObject(data)
    {
      Sphere::org = org;
      Sphere::r = r;
      r2 = r * r;
      Sphere::color = color;
      tex = NULL;
    }
    Sphere(SurfaceData data, Vector org, double r, Texture *tex):
    GObject(data)
    {
      Sphere::org = org;
      Sphere::r = r;
      r2 = r * r;
      Sphere::tex = tex;
    }
    bool Intersect(Ray ray, double& dist)
    {
      Vector l = org - ray.org;
      double l2oc = l & l;
      double tca = l & ray.dir;
      double t2hc = r2 - l2oc + tca * tca;
      double t2;

      if (t2hc < 0)
        return false;

      t2hc = sqrt(t2hc);

      if (tca < t2hc)
      {
        dist = tca + t2hc;
        t2 = tca - t2hc;
      }
      else
      {
        dist = tca - t2hc;
        t2 = tca + t2hc;
      }

      if (fabs(dist) < GeomThreshold)
        dist = t2;

      return dist > GeomThreshold;
    }
    Vector GetNormal(Vector p)
    {
      return (p - org) / r;
    }
    Vector GetColor(Vector p)
    {
      if (tex == NULL)
        return color;
      Vector dir = p - org;
      double y = acos((dir & Vector(0, 0, 1)) / !dir) * DEG;
      dir.z = 0;
      dir = Normalize(dir);
      double x = acos((dir & Vector(1, 0, 0)) / !dir) * DEG;
      return tex -> GetColor(x, y);
    }
};

class Cube: public GObject {
    double a, b, c;
    Vector org;
    Plane *pl[6], *last;
  public:
    Cube(SurfaceData data, Vector org, double a, double b, double c, Vector color):
    GObject(data)
    {
      Cube::org = org;
      Cube::a = a;
      Cube::b = b;
      Cube::c = c;
      pl[0] = new XYPlane(data, org.z - c, color);
      pl[1] = new XYPlane(data, org.z + c, color);
      pl[2] = new YZPlane(data, org.x - b, color);
      pl[3] = new YZPlane(data, org.x + b, color);
      pl[4] = new ZXPlane(data, org.y - a, color);
      pl[5] = new ZXPlane(data, org.y + a, color);
    }
    Cube(SurfaceData data, Vector org, double a, double b, double c, Texture *tex):
    GObject(data)
    {
      Cube::org = org;
      Cube::a = a;
      Cube::b = b;
      Cube::c = c;
      pl[0] = new XYPlane(data, org.z - c, tex);
      pl[1] = new XYPlane(data, org.z + c, tex);
      pl[2] = new YZPlane(data, org.x - b, tex);
      pl[3] = new YZPlane(data, org.x + b, tex);
      pl[4] = new ZXPlane(data, org.y - a, tex);
      pl[5] = new ZXPlane(data, org.y + a, tex);
    }
    ~Cube()
    {
      for (int i = 0; i < 6; i++)
        delete pl[i];
    }
    bool Intersect(Ray ray, double& dist)
    {
      double d[6], min = 1.7E+308;
      bool b_[6];
      Vector p[6];

      for (int i = 0; i < 6; i++)
      {
        b_[i] = pl[i] -> Intersect(ray, d[i]);
        p[i] = ray.Point(d[i]) - org;
        if (p[i].x > a || p[i].x < -a || p[i].y > b || p[i].y < -b || p[i].z > c || p[i].z < -c)
          b_[i] = false;
      }

      for (int i = 0; i < 6; i++)
        if (b_[i])
          if (d[i] < min)
          {
            min = d[i];
            last = pl[i];
          }

      if (min < 1.7E+308)
      {
        dist = min;
        return true;
      }

      return false;
    }
    Vector GetNormal(Vector p)
    {
      return last -> GetNormal(p);
    }
    Vector GetColor(Vector p)
    {
      return last -> GetColor(p);
    }
};

#endif