#ifndef _SCENE_H_
#define _SCENE_H_
#include <windows.h>
#include "vector.h"
#include "matrix.h"
#include "surface.h"
#include "objects.h"
#include "lights.h"

class Scene {
    Medium med, old_med;
    int cur_ray, max_rays;
    long width, height;
    double last_dist, percent;
    double fog_start, fog_end, fog_density, wave_amp, wave_size;
    char *buf;
    Vector bgcolor, ambient, fog_color;
    GObject *beg_obj, *end_obj;
    Light *beg_light, *end_light;
    Vector Trace(Ray, GObject *, Medium, GObject *);
    GObject *Intersect(Ray, GObject *, double&);
    bool fog_enabled;
  public:
    Scene(Vector, Vector, Medium);
    ~Scene();
    void AddObject(GObject *obj);
    void AddLight(Light *light);
    void RenderScene(int, int, double, double, Vector, Vector, Vector, int, char *);
    void EnableFog();
    void DisableFog();
    void SetFog(double, double, Vector, double, double, double);
    double GetCompletedVal()
    {
      return percent * 100;
    }
};

Scene::Scene(Vector bgcolor, Vector ambient, Medium med)
{
  cur_ray = 0;
  beg_obj = end_obj = NULL;
  beg_light = end_light = NULL;
  Scene::bgcolor = bgcolor;
  Scene::ambient = ambient;
  Scene::med = med;
  old_med = med;
  fog_enabled = false;
  fog_color = 0;
  fog_start = 0;
  fog_end = 0;
  fog_density = 0;
  wave_amp = 0;
  wave_size = 0;
}

Scene::~Scene()
{
  GObject *obj = beg_obj;
  Light *light = beg_light;

  while (obj != NULL)
  {
    beg_obj = obj -> next;
    delete obj;
    obj = beg_obj;
  }

  while (light != NULL)
  {
    beg_light = light -> next;
    delete light;
    light = beg_light;
  }
}

void Scene::AddObject(GObject *obj)
{
  if (beg_obj == NULL)
    beg_obj = obj;
  else
    end_obj -> next = obj;
  end_obj = obj;
}

void Scene::AddLight(Light *light)
{
  if (beg_light == NULL)
    beg_light = light;
  else
    end_light -> next = light;
  end_light = light;
}

void Scene::EnableFog()
{
  fog_enabled = true;
}

void Scene::DisableFog()
{
  fog_enabled = false;
}

void Scene::SetFog(double start, double end, Vector color, double density, double amp, double size)
{
  fog_color = color;
  fog_start = start;
  fog_end = end;
  fog_density = density;
  wave_amp = amp;
  wave_size = size;
}

void Scene::RenderScene(int width, int height, double nx, double ny, Vector camera_org, Vector camera_dir, Vector up_dir, int max_rays, char *buf)
{
  Scene::max_rays = max_rays;
  Vector color, vx, vy;
  double x, y, x_step = 2.0 * nx / width, y_step = 2.0 * ny / height;
  Ray ray;
  int i, j;
  long p = 0;

  vx = Normalize(up_dir ^ camera_dir);
  vy = Normalize(camera_dir ^ vx);
  ray.org = camera_org;
  percent = 0;

  for (i = 0, y = ny; i < height; i++, y -= y_step)
  {
    for (j = 0, x = nx; j < width; j++, x -= x_step)
    {
      ray.dir = Normalize(camera_dir + vx * x + vy * y);
      color = Trace(ray, NULL, med, NULL);
      color = Clip(color);
      buf[p] = color.z * 255;
      buf[p + 1] = color.y * 255;
      buf[p + 2] = color.x * 255;
      p += 3;
    }
    percent = (double)i / height;
  }
}

Vector Scene::Trace(Ray ray, GObject *ignore, Medium cur_med, GObject *cur_obj)
{
  Medium new_med;
  Vector color(bgcolor), reflect, refract, sum(0), ldir, point, n, v, sum2(0), c;
  GObject *o, *obj;
  Light *light = beg_light;
  double dist, t;
  Ray r;
  SurfaceData data;
  cur_ray++;

  if ((obj = Intersect(ray, ignore, dist)) != NULL)
  {
    point = ray.Point(dist);
    data = obj -> data;
    n = obj -> GetNormal(point);
    r.org = point;

    if (cur_ray < max_rays && data.Kr > 0 && cur_obj != obj)
    {
      r.dir = ray.dir - n * (2 * (ray.dir & n));
      reflect = Trace(r, obj, cur_med, cur_obj);
    }
    else
      reflect = 0;

    if (cur_ray < max_rays && data.Kt > 0)
    {
      double c;
      if (cur_obj == obj)
      {
        c = ray.dir & n;
        new_med = old_med;
      }
      else
      {
        c = -ray.dir & n;
        new_med = data.med;
      }
      old_med = cur_med;
      double et = cur_med.refr / new_med.refr;
      r.dir = et * ray.dir + (et * c - sqrt(1 + et * et * (c * c - 1))) * n;
      refract = Trace(r, NULL, new_med, obj);
    }
    else
      refract = 0;

    if (data.Kd > 0 || data.Ks > 0)
    {
      while (light != NULL)
      {
        v = light -> org - point;
        ldir = Normalize(v);
        r.dir = ldir;
        o = Intersect(r, obj, t);
        if (o == NULL || t > !v)
        {
          c = light -> GetColor(cur_med.betta, v, n);
          sum += c;
          sum2 += c * fabs(pow(n & Normalize(ldir - ray.dir), data.p));
        }
        light = light -> next;
      }
    }

    color = data.Ka * ambient +
            data.Kd * obj -> GetColor(point) * sum +
            data.Ks * sum2 +
            data.Kr * reflect * exp(-cur_med.betta * last_dist) +
            data.Kt * refract * exp(-cur_med.betta * last_dist);

    if (fog_enabled)
    {
      double d, f = fog_end * (1 + sin(point.x * wave_size) * sin(point.y * wave_size) * wave_amp);
      if (point.z <= fog_start)
        d = 0;
      else if (point.z >= f)
        d = 1;
      else
      {
        double v = 1 - (point.z - fog_start) / fabs(f - fog_start);
        d = exp(-v * fog_density * !(ray.org - point));
      }
      color = d * color + fog_color * (1 - d);
    }
  }
  else
  {
    color = bgcolor;
    dist = 1.7E+308;
  }

  cur_ray--;
  last_dist = dist;
  return color;
}

GObject *Scene::Intersect(Ray ray, GObject *ignore, double& dist)
{
  GObject *obj = beg_obj, *o = NULL;
  double min = 1.7E+308;

  while (obj != NULL)
  {
    if (obj != ignore)
      if (obj -> Intersect(ray, dist))
        if (dist < min)
        {
          min = dist;
          o = obj;
        }
    obj = obj -> next;
  }

  dist = min;
  return o;
}

#endif
